@@ -1,5 +1,5 @@
 #!/usr/bin/env python
-# Copyright (C) 2015, 2016 IBM Corp. All rights reserved.
+# Copyright (C) 2015, 2016, 2017 IBM Corp. All rights reserved.
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -19,13 +19,16 @@ instance.
 import base64
 import json
 import posixpath
-import requests
 
 from ._2to3 import bytes_, unicode_
 from .database import CloudantDatabase, CouchDatabase
 from .feed import Feed, InfiniteFeed
-from .error import CloudantException, CloudantArgumentError
-from ._common_util import USER_AGENT, append_response_error_content
+from .error import CloudantArgumentError, CloudantClientException
+from ._common_util import (
+    USER_AGENT,
+    append_response_error_content,
+    InfiniteSession,
+    ClientSession)
 
 
 class CouchDB(dict):
@@ -40,12 +43,27 @@ class CouchDB(dict):
 
     :param str user: Username used to connect to CouchDB.
     :param str auth_token: Authentication token used to connect to CouchDB.
-    :param str url: URL for CouchDB server.
     :param bool admin_party: Setting to allow the use of Admin Party mode in
         CouchDB.  Defaults to ``False``.
+    :param str url: URL for CouchDB server.
     :param str encoder: Optional json Encoder object used to encode
         documents for storage.  Defaults to json.JSONEncoder.
-    :param requests.HTTPAdapter adapter: Optional adapter to use for configuring requests.
+    :param requests.HTTPAdapter adapter: Optional adapter to use for
+        configuring requests.
+    :param bool connect: Keyword argument, if set to True performs the call to
+        connect as part of client construction.  Default is False.
+    :param bool auto_renew: Keyword argument, if set to True performs
+        automatic renewal of expired session authentication settings.
+        Default is False.
+    :param float timeout: Timeout in seconds (use float for milliseconds, for
+        example 0.1 for 100 ms) for connecting to and reading bytes from the
+        server.  If a single value is provided it will be applied to both the
+        connect and read timeouts.  To specify different values for each timeout
+        use a tuple.  For example, a 10 second connect timeout and a 1 minute
+        read timeout would be (10, 60).  This follows the same behaviour as the
+        `Requests library timeout argument
+        <http://docs.python-requests.org/en/master/user/quickstart/#timeouts>`_.
+        but will apply to every request made using this client.
     """
     _DATABASE_CLASS = CouchDatabase
 
@@ -59,22 +77,41 @@ class CouchDB(dict):
         self.admin_party = admin_party
         self.encoder = kwargs.get('encoder') or json.JSONEncoder
         self.adapter = kwargs.get('adapter')
+        self._timeout = kwargs.get('timeout', None)
         self.r_session = None
+        self._auto_renew = kwargs.get('auto_renew', False)
+        connect_to_couch = kwargs.get('connect', False)
+        if connect_to_couch and self._DATABASE_CLASS == CouchDatabase:
+            self.connect()
 
     def connect(self):
         """
         Starts up an authentication session for the client using cookie
-        authentication.
+        authentication if necessary.
         """
-        self.r_session = requests.Session()
+        if self.r_session:
+            return
+
+        if self._auto_renew and not self.admin_party:
+            self.r_session = InfiniteSession(
+                self._user,
+                self._auth_token,
+                self.server_url,
+                timeout=self._timeout
+            )
+        else:
+            self.r_session = ClientSession(
+                self._user,
+                self._auth_token,
+                self.server_url,
+                timeout=self._timeout
+            )
         # If a Transport Adapter was supplied add it to the session
         if self.adapter is not None:
             self.r_session.mount(self.server_url, self.adapter)
         if self._client_user_header is not None:
             self.r_session.headers.update(self._client_user_header)
-        if not self.admin_party:
-            self.r_session.auth = (self._user, self._auth_token)
-            self.session_login(self._user, self._auth_token)
+        self.session_login(self._user, self._auth_token)
         self._client_session = self.session()
         # Utilize an event hook to append to the response message
         # using :func:`~cloudant.common_util.append_response_error_content`
@@ -176,37 +213,34 @@ class CouchDB(dict):
         Creates a new database on the remote server with the name provided
         and adds the new database object to the client's locally cached
         dictionary before returning it to the caller.  The method will
-        optionally throw a CloudantException if the database exists remotely.
+        optionally throw a CloudantClientException if the database
+        exists remotely.
 
         :param str dbname: Name used to create the database.
         :param bool throw_on_exists: Boolean flag dictating whether or
-            not to throw a CloudantException when attempting to create a
-            database that already exists.
+            not to throw a CloudantClientException when attempting to
+            create a database that already exists.
 
         :returns: The newly created database object
         """
         new_db = self._DATABASE_CLASS(self, dbname)
         if new_db.exists():
             if kwargs.get('throw_on_exists', True):
-                raise CloudantException(
-                    "Database {0} already exists".format(dbname)
-                )
+                raise CloudantClientException(409, dbname)
         new_db.create()
         super(CouchDB, self).__setitem__(dbname, new_db)
         return new_db
 
     def delete_database(self, dbname):
         """
-        Removes the named database remotely and locally. The method will throw a
-        CloudantException if the database does not exist.
+        Removes the named database remotely and locally. The method will throw
+        a CloudantClientException if the database does not exist.
 
         :param str dbname: Name of the database to delete.
         """
         db = self._DATABASE_CLASS(self, dbname)
         if not db.exists():
-            raise CloudantException(
-                "Database {0} does not exist".format(dbname)
-            )
+            raise CloudantClientException(404, dbname)
         db.delete()
         if dbname in list(self.keys()):
             super(CouchDB, self).__delitem__(dbname)
@@ -356,8 +390,7 @@ class CouchDB(dict):
             create the database remotely or not.  Defaults to False.
         """
         if not isinstance(value, self._DATABASE_CLASS):
-            msg = "Value must be set to a Database object"
-            raise CloudantException(msg)
+            raise CloudantClientException(101, type(value).__name__)
         if remote and not value.exists():
             value.create()
         super(CouchDB, self).__setitem__(key, value)
@@ -392,7 +425,6 @@ class Cloudant(CouchDB):
 
     def __init__(self, cloudant_user, auth_token, **kwargs):
         super(Cloudant, self).__init__(cloudant_user, auth_token, **kwargs)
-
         self._client_user_header = {'User-Agent': USER_AGENT}
         account = kwargs.get('account')
         url = kwargs.get('url')
@@ -405,7 +437,10 @@ class Cloudant(CouchDB):
                 self._client_user_header['X-Cloudant-User'] = x_cloudant_user
 
         if self.server_url is None:
-            raise CloudantException('You must provide a url or an account.')
+            raise CloudantClientException(102)
+
+        if kwargs.get('connect', False):
+            self.connect()
 
     def db_updates(self, raw_data=False, **kwargs):
         """
@@ -513,7 +548,7 @@ class Cloudant(CouchDB):
             between 1 and 12. Optional parameter. Defaults to None.
             If used, it must be accompanied by ``year``.
         """
-        err = None
+        err = False
         if year is None and month is None:
             resp = self.r_session.get(endpoint)
         else:
@@ -524,14 +559,12 @@ class Cloudant(CouchDB):
                             endpoint, str(int(year)), str(int(month)))
                     )
                 else:
-                    err = ('Invalid year and/or month supplied.  '
-                           'Found: year - {0}, month - {1}').format(year, month)
+                    err = True
             except (ValueError, TypeError):
-                err = ('Invalid year and/or month supplied.  '
-                       'Found: year - {0}, month - {1}').format(year, month)
+                err = True
 
         if err:
-            raise CloudantArgumentError(err)
+            raise CloudantArgumentError(101, year, month)
         else:
             resp.raise_for_status()
             return resp.json()
@@ -721,7 +754,7 @@ class Cloudant(CouchDB):
         )
         resp = self.r_session.put(
             endpoint,
-            data=json.dumps(config),
+            data=json.dumps(config, cls=self.encoder),
             headers={'Content-Type': 'application/json'}
         )
         resp.raise_for_status()
