@@ -30,10 +30,11 @@ from ._common_util import (
     get_docs)
 from .document import Document
 from .design_document import DesignDocument
+from .security_document import SecurityDocument
 from .view import View
 from .index import Index, TextIndex, SpecialIndex
 from .query import Query
-from .error import CloudantException, CloudantArgumentError
+from .error import CloudantArgumentError, CloudantDatabaseException
 from .result import Result, QueryResult
 from .feed import Feed, InfiniteFeed
 
@@ -54,10 +55,18 @@ class CouchDatabase(dict):
         self.client = client
         self._database_host = client.server_url
         self.database_name = database_name
-        self.r_session = client.r_session
         self._fetch_limit = fetch_limit
         self.result = Result(self.all_docs)
 
+    @property
+    def r_session(self):
+        """
+        Returns the ``r_session`` from the client instance used by the database.
+
+        :returns: Client ``r_session``
+        """
+        return self.client.r_session
+
     @property
     def admin_party(self):
         """
@@ -101,7 +110,10 @@ class CouchDatabase(dict):
 
         :returns: Boolean True if the database exists, False otherwise
         """
-        resp = self.r_session.get(self.database_url)
+        resp = self.r_session.head(self.database_url)
+        if resp.status_code not in [200, 404]:
+            resp.raise_for_status()
+
         return resp.status_code == 200
 
     def metadata(self):
@@ -146,9 +158,7 @@ class CouchDatabase(dict):
         else:
             doc = Document(self, docid)
         if throw_on_exists and doc.exists():
-            raise CloudantException(
-                'Error - Document with id {0} already exists.'.format(docid)
-            )
+            raise CloudantDatabaseException(409, docid)
         doc.update(data)
         doc.create()
         super(CouchDatabase, self).__setitem__(doc['_id'], doc)
@@ -215,6 +225,19 @@ class CouchDatabase(dict):
 
         return ddoc
 
+    def get_security_document(self):
+        """
+        Retrieves the database security document as a SecurityDocument object.
+        The returned object is useful for viewing as well as updating the
+        the database's security document.
+
+        :returns: A SecurityDocument instance representing the database
+            security document
+        """
+        sdoc = SecurityDocument(self)
+        sdoc.fetch()
+        return sdoc
+
     def get_view_result(self, ddoc_id, view_name, raw_result=False, **kwargs):
         """
         Retrieves the view result based on the design document and view name.
@@ -321,11 +344,8 @@ class CouchDatabase(dict):
         if resp.status_code == 201 or resp.status_code == 202:
             return self
 
-        raise CloudantException(
-            "Unable to create database {0}: Reason: {1}".format(
-                self.database_url, resp.text
-            ),
-            code=resp.status_code
+        raise CloudantDatabaseException(
+            resp.status_code, self.database_url, resp.text
         )
 
     def delete(self):
@@ -649,7 +669,7 @@ class CouchDatabase(dict):
         headers = {'Content-Type': 'application/json'}
         resp = self.r_session.post(
             url,
-            data=json.dumps(data),
+            data=json.dumps(data, cls=self.client.encoder),
             headers=headers
         )
         resp.raise_for_status()
@@ -673,7 +693,7 @@ class CouchDatabase(dict):
         resp = self.r_session.post(
             url,
             headers={'Content-Type': 'application/json'},
-            data=json.dumps(data)
+            data=json.dumps(data, cls=self.client.encoder)
         )
         resp.raise_for_status()
 
@@ -702,7 +722,7 @@ class CouchDatabase(dict):
         resp = self.r_session.post(
             url,
             headers={'Content-Type': 'application/json'},
-            data=json.dumps(data)
+            data=json.dumps(data, cls=self.client.encoder)
         )
         resp.raise_for_status()
 
@@ -722,10 +742,7 @@ class CouchDatabase(dict):
         try:
             ret = int(resp.text)
         except ValueError:
-            resp.status_code = 400
-            raise CloudantException(
-                'Error - Invalid Response Value: {}'.format(resp.json())
-            )
+            raise CloudantDatabaseException(400, resp.json())
 
         return ret
 
@@ -741,7 +758,7 @@ class CouchDatabase(dict):
         """
         url = posixpath.join(self.database_url, '_revs_limit')
 
-        resp = self.r_session.put(url, data=json.dumps(limit))
+        resp = self.r_session.put(url, data=json.dumps(limit, self.client.encoder))
         resp.raise_for_status()
 
         return resp.json()
@@ -891,6 +908,7 @@ class CouchDatabase(dict):
             resp = self.r_session.post(
                 '/'.join([ddoc.document_url, '_update', handler_name]),
                 params=params, data=data)
+        resp.raise_for_status()
         return resp.text
 
 class CloudantDatabase(CouchDatabase):
@@ -918,11 +936,9 @@ class CloudantDatabase(CouchDatabase):
         containing information about the users that the database
         is shared with.
 
-        :returns: Security document in JSON format
+        :returns: Security document as a ``dict``
         """
-        resp = self.r_session.get(self.security_url)
-        resp.raise_for_status()
-        return resp.json()
+        return dict(self.get_security_document())
 
     @property
     def security_url(self):
@@ -970,16 +986,13 @@ class CloudantDatabase(CouchDatabase):
             perms = list(set(roles))
 
         if not perms:
-            msg = (
-                'Invalid role(s) provided: {0}.  Valid roles are: {1}.'
-            ).format(roles, valid_roles)
-            raise CloudantArgumentError(msg)
+            raise CloudantArgumentError(102, roles, valid_roles)
 
         data[username] = perms
         doc['cloudant'] = data
         resp = self.r_session.put(
             self.security_url,
-            data=json.dumps(doc),
+            data=json.dumps(doc, cls=self.client.encoder),
             headers={'Content-Type': 'application/json'}
         )
         resp.raise_for_status()
@@ -1004,7 +1017,7 @@ class CloudantDatabase(CouchDatabase):
         doc['cloudant'] = data
         resp = self.r_session.put(
             self.security_url,
-            data=json.dumps(doc),
+            data=json.dumps(doc, cls=self.client.encoder),
             headers={'Content-Type': 'application/json'}
         )
         resp.raise_for_status()
@@ -1066,7 +1079,7 @@ class CloudantDatabase(CouchDatabase):
                     **data.get('def', {})
                 ))
             else:
-                raise CloudantException('Unexpected index content: {0} found.')
+                raise CloudantDatabaseException(101, data.get('type'))
         return indexes
 
     def create_query_index(
@@ -1115,11 +1128,7 @@ class CloudantDatabase(CouchDatabase):
         elif index_type == TEXT_INDEX_TYPE:
             index = TextIndex(self, design_document_id, index_name, **kwargs)
         else:
-            msg = (
-                'Invalid index type: {0}.  '
-                'Index type must be either \"json\" or \"text\"'
-            ).format(index_type)
-            raise CloudantArgumentError(msg)
+            raise CloudantArgumentError(103, index_type)
         index.create()
         return index
 
@@ -1139,11 +1148,7 @@ class CloudantDatabase(CouchDatabase):
         elif index_type == TEXT_INDEX_TYPE:
             index = TextIndex(self, design_document_id, index_name)
         else:
-            msg = (
-                'Invalid index type: {0}.  '
-                'Index type must be either \"json\" or \"text\"'
-            ).format(index_type)
-            raise CloudantArgumentError(msg)
+            raise CloudantArgumentError(103, index_type)
         index.delete()
 
     def get_query_result(self, selector, fields=None, raw_result=False,
@@ -1330,20 +1335,14 @@ class CloudantDatabase(CouchDatabase):
         param_query = query_params.get('query')
         # Either q or query parameter is required
         if bool(param_q) == bool(param_query):
-            raise CloudantArgumentError(
-                'A single query/q parameter is required. '
-                'Found: {0}'.format(query_params))
+            raise CloudantArgumentError(104, query_params)
 
         # Validate query arguments and values
         for key, val in iteritems_(query_params):
             if key not in list(SEARCH_INDEX_ARGS.keys()):
-                msg = 'Invalid argument: {0}'.format(key)
-                raise CloudantArgumentError(msg)
+                raise CloudantArgumentError(105, key)
             if not isinstance(val, SEARCH_INDEX_ARGS[key]):
-                msg = (
-                    'Argument {0} is not an instance of expected type: {1}'
-                ).format(key, SEARCH_INDEX_ARGS[key])
-                raise CloudantArgumentError(msg)
+                raise CloudantArgumentError(106, key, SEARCH_INDEX_ARGS[key])
         # Execute query search
         headers = {'Content-Type': 'application/json'}
         ddoc = DesignDocument(self, ddoc_id)
